from __future__ import annotations

import sys

import pytest
import torch

from rebasin import MergeMany
from tests.fixtures.models import MLP


def test_merge_many__get_mean_model() -> None:
    models = [MLP(10, 10), MLP(10, 10), MLP(10, 10)]
    x = torch.randn(10)
    merge_many = MergeMany(models=models, working_model=MLP(10, 10), input_data=x)
    mean_model = merge_many._get_mean_model()

    manual_mean_model = MLP(10, 10)
    manual_mean_model.load_state_dict(models[0].state_dict())
    for model in models[1:]:
        for param, mean_param in zip(
                model.parameters(), manual_mean_model.parameters()
        ):
            mean_param.data.add_(param.data)

    for param in manual_mean_model.parameters():
        param.data.div_(3.0)

    for param, mean_param in zip(
            mean_model.parameters(), manual_mean_model.parameters()
    ):
        assert torch.allclose(param, mean_param)

    # Test with except_index.
    mean_model = merge_many._get_mean_model(except_index=0)
    manual_mean_model = MLP(10, 10)
    manual_mean_model.load_state_dict(models[1].state_dict())
    for param, mean_param in zip(
            models[2].parameters(), manual_mean_model.parameters()
    ):
        mean_param.data.add_(param.data)

    for param in manual_mean_model.parameters():
        param.data.div_(2.0)

    for param, mean_param in zip(
            mean_model.parameters(), manual_mean_model.parameters()
    ):
        assert torch.allclose(param, mean_param)


def test_merge_many() -> None:
    models = [MLP(5, 5), MLP(5, 5), MLP(5, 5)]
    x = torch.randn(5)
    merge_many = MergeMany(models=models, working_model=MLP(5, 5), input_data=x)

    # I cannot write any proper tests here, because I would need data for that,
    #   and I don't want to download any for testing.
    # I will just test that the code runs without errors.
    merged_model = merge_many.run()


@pytest.mark.skipif(
    "--full-suite" not in sys.argv,
    reason="This test takes a long time to run."
)
def test_merge_many_large_mlp() -> None:
    models = [MLP(50, 50), MLP(50, 50), MLP(50, 50)]
    x = torch.randn(50)
    merge_many = MergeMany(models=models, working_model=MLP(50, 50), input_data=x)

    # I cannot write any proper tests here, because I would need data for that,
    #   and I don't want to download any for testing.
    # I will just test that the code runs without errors.
    _ = merge_many.run()



from __future__ import annotations

import copy
import logging
from collections.abc import Sequence
from pathlib import Path
from typing import Any

import torch
from torch import nn
from torch.utils.data import DataLoader
from tqdm import tqdm

from rebasin.utils import parse_logging_level, recalculate_batch_norms


class Interpolation:
    """
    Interpolate between models.

    The base class for all interpolation classes.
    **This class is not meant to be used directly.**

    Args:
        models:
            The models to interpolate between.
            This can be multiple models (see explanation above).

            *Type:* :code:`Sequence[nn.Module]`

        eval_fn:
            The function to evaluate the interpolated models with.
            Usually, this simply calculates the loss for every batch in a
            validation or test dataloader, but it can be anything,
            including accuracy, or perplexity, etc.
            By using a :code:`eval_fn`, the user has maximum control.

            Its signature must be as follows:

                >>> def eval_fn(
                >>>     model: nn.Module, device: torch.device | str | None
                >>> ) -> float:
                >>>     ...

            The :code:`model` argument is necessary
            because different models will have to be
            evaluated. The :code:`device` argument is necessary because the different
            models can be evaluated on different devices
            (see the :code:`devices` argument).

            If the default :code:`eval_fn` is used, then the best model will always be
            :code:`models[0]`, and all metrics will be :code:`0.0`.
            In this case, it is recommended to provide a :code:`savedir`
            so that the interpolated models are saved and can be evaluated manually.

            *Type:* Callable[[nn.Module, torch.device | str | None], float]

            *Default:* :code:`lambda model, device: 0.0`

        eval_mode:
            The mode to use for evaluation. If "min", the model with the lowest
            metric is chosen, if "max", the model with the highest metric is chosen.

            *Type:* :code:`str`

            *Default:* :code:`"min"`

        train_dataloader:
            The dataloader to use for training the models.
            If this is given, then the :class:`Interpolation` subclass will check if
            the model has :code:`BatchNorm` layers and if so, will
            recalculate their :code:`running_mean` and :code:`running_var`
            statistics using the training data.

            This is unfortunately strongly recommended if your model includes
            these statistics.

            *Type:* :class:`torch.utils.data.DataLoader | None`

            *Default:* :code:`None`

        dataset_percentage:
            The percentage of the dataset to use
            for recalculating the BatchNorm statistics.
            Must be between :code:`0.0` and :code:`1.1`.

            *Type:* :code:`float`

            *Default:* :code:`1.0`

        dataset_iterations:
            An alternative to :code:`dataset_percentage`.
            If it is > 0, then this will be the maximum number of batches
            to use from the training dataloader for recalculating
            the BatchNorm statistics.

            *Type:* :code:`int`

            *Default:* :code:`-1`

        input_indices:
            If a training dataloader is given to :code:`train_dataloader`,
            and the model has :code:`BatchNorm` layers, then the
            :code:`running_mean` and :code:`running_var` statistics
            are recalculated using the training data.
            However, :class:`Interpolation` does not know which output from
            the dataloader is the input to the model. Sometimes,
            a model takes several inputs, sometimes, and input and a target
            are given by the dataloader, etc.
            To solve this, the argument :code:`input_indices` can be used
            to provide the indices of the inputs to the model.

            Can be an integer or a sequence of integers.

            The outputs of the training dataloader corresponding to the
            indices given here will be used as inputs to the models to
            recalculate the :code:`running_mean` and :code:`running_var`

            *Type:* :code:`Sequence[int] | int`

            *Default:* :code:`0`

        devices:
            Models may need a large amount of GPU-memory.
            To avoid running out of memory,
            the argument :code:`devices` can be used to specify which device
            each of the given models in :code:`models` is on.
            As each model is evaluated using the function given in :code:`eval_fn`,
            the corresponding device is passed to :code:`eval_fn` as well.

            *Type:* :code:`Sequence[torch.device | str | None] | None`

            *Default:* :code:`None`

        device_interp:
            The device to use for the interpolation.
            If this is :code:`None`, no parameter will be moved
            to a different device for interpolation, and the
            interpolated model will be created on CPU.
            Again, this argument is useful for saving on memory.

            *Type:* :code:`torch.device | str | None`

            *Default:* :code:`None`

        savedir:
            The directory to save the models in.
            If :code:`None`, the models are not saved.
            Otherwise, the interpolated models are saved in the given directory.

            Type: :class:`pathlib.Path` or :class:`str` or :code:`None`.

            Default: :code:`None`.

        logging_level:
            The logging level to use.

            Can be any of the following:

            - :code:`logging.DEBUG` / :code:`"DEBUG"` / :code:`"debug"` / :code:`10`
            - :code:`logging.INFO` / :code:`"INFO"` / :code:`"info"` / :code:`20`
            - :code:`logging.WARNING` / :code:`"WARNING"`
              / :code:`"warning"` / :code:`30`
            - :code:`logging.WARN`  / :code:`"WARN"` / :code:`"warn"` / :code:`30`
            - :code:`logging.ERROR` / :code:`"ERROR"` / :code:`"error"` / :code:`40`
            - :code:`logging.CRITICAL` / :code:`"CRITICAL"`
              / :code:`"critical"` / :code:`50`
            - :code:`logging.FATAL` / :code:`"FATAL"` / :code:`"fatal"` / :code:`50`

            Type: :code:`int | str`.

            Default: :code:`logging.ERROR`.
    """
    def __init__(
            self,
            models: Sequence[torch.nn.Module],
            eval_fn: Any = lambda model, device: 0.0,
            eval_mode: str = "min",
            train_dataloader: DataLoader[Any] | None = None,
            dataset_percentage: float = 1.0,
            dataset_iterations: int = -1,
            input_indices: Sequence[int] | int = 0,
            devices: Sequence[torch.device | str] | None = None,
            device_interp: torch.device | str | None = None,
            savedir: Path | str | None = None,
            logging_level: int | str = "ERROR"
    ) -> None:
        self._sanity_checks(
            models,
            eval_fn,
            eval_mode,
            train_dataloader,
            dataset_percentage,
            dataset_iterations,
            devices,
            device_interp,
            input_indices,
            savedir,
        )
        self.logging_level = parse_logging_level(logging_level)
        if self.logging_level <= logging.INFO:
            print("Setting up interpolation...")

        if savedir is not None and isinstance(savedir, str):
            savedir = Path(savedir)

        self.models = models
        self.eval_fn = eval_fn
        self.eval_mode = eval_mode
        self.idx_fn = torch.argmax if eval_mode == "max" else torch.argmin
        self.train_dataloader = train_dataloader
        self.dataset_percentage = dataset_percentage
        self.dataset_iterations = dataset_iterations
        self.devices: Sequence[torch.device | str | None] = (
            devices if devices is not None
            else [None] * len(models)  # type: ignore[list-item]
        )
        self.device_interp = device_interp
        self.input_indices = input_indices
        self.savedir = savedir

        self.metrics_interpolated: list[float] = []
        self.best_metric: float = float("inf") if eval_mode == "min" else -float("inf")

    def interpolate(self, steps: int = 20, savedir: str | Path | None = None) -> None:
        """Interpolate between the models and save the best one or all."""

    @staticmethod
    def _sanity_checks(
            models: Sequence[torch.nn.Module],
            eval_fn: Any,
            eval_mode: str,
            train_dataloader: DataLoader[Any] | None,
            dataset_percentage: float,
            dataset_iterations: int,
            devices: Sequence[torch.device | str] | None,
            device_interp: torch.device | str | None,
            input_indices: Sequence[int] | int,
            savedir: Path | str | None,
    ) -> None:
        assert isinstance(models, Sequence), "Models must be a sequence"
        assert all(isinstance(model, nn.Module) for model in models), \
            "All models must be a subclass of nn.Module"

        assert callable(eval_fn), "Eval function must be callable"
        assert isinstance(eval_mode, str)
        assert eval_mode in ["min", "max"], "Eval mode must be 'min' or 'max'"

        assert isinstance(train_dataloader, (DataLoader, type(None)))
        assert isinstance(dataset_percentage, float), "dataset_percentage must be float"
        assert 0.0 <= dataset_percentage <= 1.0, "dataset_percentage must be in [0, 1]"
        assert isinstance(dataset_iterations, int), "dataset_iterations must be int"

        if devices is not None:
            assert isinstance(devices, Sequence)
            assert all(
                isinstance(device, (str, torch.device)) for device in devices
            )
            assert len(devices) == len(models)

        assert isinstance(device_interp, (str, torch.device, type(None)))

        assert isinstance(input_indices, (int, Sequence))
        if isinstance(input_indices, Sequence):
            assert all(isinstance(i, int) for i in input_indices)

        assert isinstance(savedir, (Path, str, type(None)))

        assert len(models) > 1, "Need at least two models to interpolate between"


class LerpSimple(Interpolation):
    r"""
    Linear interpolation between models.

    For two models and `s` steps, the interpolation looks like this
    (using :math:`m_a` to denote model number :math:`a`,
    and :math:`i_{ab,c}` to denote interpolation step `c`
    between models number `a` and `b`):

    |

    .. math::

        m_1 - i_{12,1} - ... - i_{12,s} - m_2

    |

    For `n` models, the same is repeated between every other pair of models,
    meaning:

    |

    .. math::
        m_1 - i_{12,1} - ... - i_{12,s} - m_2
        - i_{23,1} - ... - i_{23,s} - m_3 - ... - m_n

    |

    This means that the total number of interpolations is :math:`s \cdot (n - 1)`.

    For args, see :class:`Interpolation`.
    """

    def interpolate(self, steps: int = 20, savedir: str | Path | None = None) -> None:
        r"""
        Interpolate between the models and save the interpolated models
        if :code:`savedir` is given.

        Args:
            steps:
                The number of steps to take between each pair of models.

            savedir:
                The directory to save the models in. Overwrites the :code:`savedir`
                argument passed to the constructor if not None.
                If both are :code:`None`, the interpolated models are not saved.

                If :code:`not None`, the following naming schema is used for the files:

                :code:`"interp_models_{m_i}_{m_i+1}_perc_{percentage}.pt"`
                if the model is interpolated.
                :math:`m_i` and :math:`m_{i+1}` are the indices of the models
                between which the interpolation is done.
                `percentage` is the percentage of the way through the interpolation.
                If it is small, the model is closer to :math:`m_i`,
                and if it is large, the model is closer to :math:`m_{i+1}`.
        """
        # SANITY CHECKS AND DEFAULT SETTINGS
        assert isinstance(savedir, (Path, str, type(None)))
        if savedir is None:
            savedir = self.savedir
        elif isinstance(savedir, str):
            savedir = Path(savedir)

        if self.logging_level <= logging.INFO:
            print("Interpolating between models...")

        # INTERPOLATION
        loop = tqdm(range(steps), disable=self.logging_level > logging.INFO)
        for step in loop:
            # Interpolate between the two models
            # (step + 1) so that it never starts at zero percent.
            # (steps + 2) so that it never ends at 100 percent.
            # This is because the start and end models already exist
            percentage = (step + 1) / (steps + 2)
            self._interpolate_step(percentage=percentage, savedir=savedir, loop=loop)

    def _interpolate_step(
            self, percentage: float, savedir: Path | None, loop: tqdm[Any]
    ) -> None:
        # Interpolate between all models
        for model_num, (model1, model2) in enumerate(
                zip(self.models[:-1], self.models[1:])
        ):
            model_interp = copy.deepcopy(model1).to(self.device_interp)

            for param1, param2, param_interp in zip(
                    model1.parameters(),
                    model2.parameters(),
                    model_interp.parameters(),
            ):
                param_interp.data = torch.lerp(
                    param1.to(self.device_interp),
                    param2.to(self.device_interp),
                    percentage
                )

            # Recalculate BatchNorm statistics
            if self.train_dataloader is not None:
                recalculate_batch_norms(
                    model_interp,
                    self.train_dataloader,
                    self.input_indices,
                    device=self.device_interp,
                    verbose=self.logging_level <= logging.INFO,
                    dataset_percentage=self.dataset_percentage,
                    loop=loop,
                    iterations=self.dataset_iterations,
                )

            # Evaluate
            metric = self.eval_fn(model_interp, self.device_interp)
            self.metrics_interpolated.append(metric)

            # Save
            filename = f"interp_models_{model_num}_{model_num+1}" \
                       f"_perc_{percentage:.3f}.pt"

            is_best = (
                metric < self.best_metric
                if self.eval_mode == "min"
                else metric > self.best_metric
            )
            if is_best:
                self.best_metric = metric
                self.best_model = model_interp

            if savedir is not None:
                torch.save(model_interp.state_dict(), savedir.joinpath(filename))



import rebasin.interpolation
import rebasin.model_info
import rebasin.modules
import rebasin.paths
import rebasin.permutation_initializer
import rebasin.type_definitions
import rebasin.utils
from rebasin.merge_many import MergeMany
from rebasin.permutation_coordinate_descent import PermutationCoordinateDescent



from __future__ import annotations

import copy
import sys
from typing import Any

import pytest
import torch
from torch import nn
from torchvision.models import (  # type: ignore[import]
    ViT_B_16_Weights,
    resnet18,
    vit_b_16,
)

from rebasin import PermutationCoordinateDescent
from rebasin.permutation_coordinate_descent import calculate_progress
from tests.fixtures.models import MLP
from tests.fixtures.utils import (
    allclose,
    model_change_percent,
    model_distance,
    model_similarity,
)


def test_calculate_progress() -> None:
    """Test the calculation of the progress."""
    cost_mat = torch.tensor([[0.0, 1.0, 2.0], [1.0, 0.0, 3.0], [2.0, 3.0, 0.0]])
    perm_old = torch.tensor([0, 1, 2])
    perm_new = torch.tensor([2, 0, 1])

    progress = calculate_progress(cost_mat, perm_old, perm_new)

    assert progress

    perm_old = torch.tensor([2, 0, 1])
    perm_new = torch.tensor([0, 1, 2])

    progress = calculate_progress(cost_mat, perm_old, perm_new)

    assert not progress


def test_long_mlp() -> None:
    """Test that :class:`PermutationCoordinateDescent`
    can handle a linear network with 100 layers
    (each with a 10x10 weight matrix).

    Handle means that it can permute the weights s.t. the output doesn't change.
    """
    model_a, model_b = MLP(10, 100), MLP(10, 100)
    model_b_orig = copy.deepcopy(model_b)
    input_data = torch.randn(10)
    y_orig = model_b(input_data)

    pcd = PermutationCoordinateDescent(model_a, model_b, input_data)
    pcd.calculate_permutations()
    pcd.apply_permutations()

    assert model_change_percent(model_b, model_b_orig) > 0.1

    y_new = model_b(input_data)
    assert torch.allclose(y_orig, y_new)


@pytest.mark.skipif(
    "--full-suite" not in sys.argv,
    reason="This test takes a long time to run. Run with --full-suite to include it.",
)
class TestPermutationCoordinateDescent:
    def test_resnet18(self) -> None:
        self.common_tests(torch.randn(1, 3, 224, 224), resnet18(), resnet18(), 1)

    def test_mlp(self) -> None:
        self.common_tests(torch.randn(50), MLP(50, 15), MLP(50, 15), 1)

    def test_vit_b_16(self) -> None:
        self.common_tests(
            torch.randn(1, 3, 224, 224),
            vit_b_16(weights=ViT_B_16_Weights.IMAGENET1K_V1),
            vit_b_16(weights=ViT_B_16_Weights.IMAGENET1K_SWAG_LINEAR_V1),
            iters=1
        )

    @staticmethod
    def common_tests(
            input_data: Any, model_a: nn.Module, model_b: nn.Module, iters: int = 10
    ) -> None:
        for _ in range(iters):
            model_a = model_a.eval()
            model_b = model_b.eval()
            model_b_old = copy.deepcopy(model_b)  # for comparison

            y_orig = model_b(input_data)

            pcd = PermutationCoordinateDescent(model_a, model_b, input_data)
            pcd.rebasin()

            assert (
                    model_distance(model_a, model_b)
                    < model_distance(model_a, model_b_old)
            )

            y_new = model_b(input_data)
            allclose(y_orig, y_new)

    @staticmethod
    def test_enforce_identity() -> None:
        model_a = resnet18()
        model_b = resnet18()
        model_b_old = copy.deepcopy(model_b)

        x = torch.randn(1, 3, 224, 224)
        pcd = PermutationCoordinateDescent(model_a, model_b, x)
        pcd.rebasin()

        assert allclose(model_b(x), model_b_old(x))
        assert (
                model_distance(model_a, model_b)
                < model_distance(model_a, model_b_old)
        )


@pytest.mark.skipif(not torch.cuda.is_available(), reason="GPU test")
class TestPCDOnGPU:
    @staticmethod
    def test_mlp() -> None:
        device_b = torch.device("cuda")
        model_a, model_b = MLP(25), MLP(25).to(device_b)
        model_b_old = copy.deepcopy(model_b)

        pcd = PermutationCoordinateDescent(
            model_a,
            model_b,
            input_data_b=torch.randn(25).to(device_b),
            input_data_a=torch.randn(25),
            device_a="cpu",
            device_b=device_b
        )
        pcd.calculate_permutations()
        pcd.apply_permutations()

        assert (
                model_similarity(model_a, model_b)
                > model_similarity(model_a, model_b_old)
        )



from __future__ import annotations

import itertools
import logging
import math
from collections.abc import Iterator, Sequence
from typing import Any

import torch
from torch import nn
from torch.utils.data import DataLoader
from tqdm import tqdm

from rebasin.model_info import ModelInfo


def recalculate_batch_norms(
        model: nn.Module,
        dataloader: DataLoader[Any],
        input_indices: int | Sequence[int],
        device: torch.device | str | None,
        verbose: bool,
        dataset_percentage: float = 1.0,
        iterations: int = -1,
        loop: tqdm[Any] | None = None,
        *forward_args: Any,
        **forward_kwargs: Any
) -> None:
    """
    Recalculate the BatchNorm statistics of a model.

    Use this after permuting the model, if your model contains BatchNorm layers.
    Returns early if the model doesn't contain any BatchNorm layers.

    Args:
        model:
            The model.
        dataloader:
            The DataLoader to use for recalculating the statistics.
            Should ideally be the training DataLoader.
        input_indices:
            Many DataLoaders return several inputs and labels.
            These can sometimes be used in unexpected ways.
            To make sure that the correct outputs of the dataloader are used
            as inputs to the model's forward pass, you can specify the indices
            at which the inputs are located, in the order that they should be
            passed to the model.
        device:
            The device on which to run the model.
        verbose:
            Whether to print progress.
        dataset_percentage:
            The percentage of the dataset to use for recalculating the statistics.
            If the batch size is such that the last batch doesn't fully fit into
            the percentage, it is still used.
        loop:
            A tqdm loop to use for progress.
            If this is used in another loop, that loop should be passed here.
            This way, instead of creating its own progress-bar and disrupting the other,
            this function will simply update the title of the other progress-bar.
        iterations:
            If > 0, the number of iterations to use for recalculating the statistics.
            Otherwise, :code:`dataset_percentage` is used.
        *forward_args:
            Any additional positional arguments to pass to the model's forward  pass.
        **forward_kwargs:
            Any additional keyword arguments to pass to the model's forward pass.
    """
    if verbose and loop is None:
        print("Recalculating BatchNorm statistics...")
    if not any(
            isinstance(m, (nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d))
            for m in model.modules()
    ):
        if verbose and loop is None:
            print("No BatchNorm layers found in model.")
        return

    training = model.training
    model.train()

    # Reset the running mean and variance
    for module in model.modules():
        if isinstance(module, (nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d)):
            module.reset_running_stats()

    max_index = (
        int(math.ceil(len(dataloader) * dataset_percentage))
        if iterations < 1
        else iterations
    )

    # Recalculate the running mean and variance
    inner_loop = (
        tqdm(range(max_index), disable=not verbose)
        if loop is None
        else range(max_index)
    )
    for i in inner_loop:
        if loop is not None:
            loop.set_description(
                f"Recalculating BatchNorm statistics ({i + 1}/{max_index})"
            )
        batch = next(iter(dataloader))
        if isinstance(batch, Sequence):
            inputs, _ = get_inputs_labels(batch, input_indices, 0, device)
        else:
            inputs = [batch]
        model(*inputs, *forward_args, **forward_kwargs)

    if not training:
        model.eval()


def get_inputs_labels(
        batch: Any,
        input_indices: int | Sequence[int] = 0,
        label_indices: int | Sequence[int] = 1,
        device: torch.device | str | int | None = None
) -> tuple[list[Any], list[Any]]:
    """
    Get the inputs and outputs from a batch.

    Args:
        batch:
            The batch.
        input_indices:
            Many DataLoaders return several inputs and labels per batch.
            These can sometimes be used in unexpected ways.
            To make sure that the correct outputs of the dataloader are used
            as inputs to the model's forward pass, you can specify the indices
            at which the inputs are located, in the order that they should be
            passed to the model.
        label_indices:
            Like `input_indices`, but for the labels.
        device:
            The device on which to run the model.

    Returns:
        The inputs and labels.
    """
    if isinstance(input_indices, int):
        input_indices = [input_indices]
    if isinstance(label_indices, int):
        label_indices = [label_indices]

    inputs = (
        [batch[i].to(device) for i in input_indices]
        if device is not None
        else [batch[i] for i in input_indices]
    )
    labels = (
        [batch[i].to(device) for i in label_indices]
        if device is not None
        else [batch[i] for i in label_indices]
    )
    return inputs, labels


def contains_parameter(
        parameters: Sequence[nn.Parameter] | Iterator[nn.Parameter],
        parameter: nn.Parameter
) -> bool:
    """
    Check if a sequence of parameters contains a parameter.

    This cannot be done via the normal `in` operator, because
    `nn.Parameter`'s `__eq__` does not work for parameters of different shapes.

    Args:
        parameters:
            The sequence of parameters.
        parameter:
            The parameter.

    Returns:
        Whether the sequence contains the parameter.
    """
    return any(param is parameter for param in parameters)


def model_info(model: nn.Module) -> ModelInfo:
    """
    Get information about a model.

    Args:
        model:
            The model.

    Returns:
        The :class:`ModelInfo` object containing the desired information.
    """
    name = model.__class__.__name__
    contains_batch_norm = any(
        isinstance(m, (nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d))
        for m in model.modules()
    )
    num_parameters = sum(p.numel() for p in model.parameters())
    num_weights = sum(1 if "weight" in n else 0 for n, _ in model.named_parameters())
    num_biases = sum(1 if "bias" in n else 0 for n, _ in model.named_parameters())
    num_elements_per_weight = [
        p.numel() for n, p in model.named_parameters() if "weight" in n
    ]
    num_elements_per_bias = [
        p.numel() for n, p in model.named_parameters() if "bias" in n
    ]

    num_permutable_elements_per_weight = []
    for pname, parameter in model.named_parameters():
        if "weight" in pname:
            shape = parameter.shape
            permutable = shape[0] if len(shape) == 1 else shape[0] * shape[1]
            num_permutable_elements_per_weight.append(permutable)

    num_elements_per_weight_mean = torch.mean(
        torch.tensor(num_elements_per_weight, dtype=torch.float)
    ).item()
    num_elements_per_weight_std = torch.std(
        torch.tensor(num_elements_per_weight, dtype=torch.float)
    ).item()
    num_elements_per_bias_mean = torch.mean(
        torch.tensor(num_elements_per_bias, dtype=torch.float)
    ).item()
    num_elements_per_bias_std = torch.std(
        torch.tensor(num_elements_per_bias, dtype=torch.float)
    ).item()

    return ModelInfo(
        name=name,
        contains_batch_norm=contains_batch_norm,
        num_parameters=num_parameters,
        num_weights=num_weights,
        num_biases=num_biases,
        num_elements_per_weight=num_elements_per_weight,
        num_permutable_elements_per_weight=num_permutable_elements_per_weight,
        num_elements_per_bias=num_elements_per_bias,
        num_elements_per_weight_mean=num_elements_per_weight_mean,
        num_elements_per_weight_std=num_elements_per_weight_std,
        num_elements_per_bias_mean=num_elements_per_bias_mean,
        num_elements_per_bias_std=num_elements_per_bias_std,
    )


def pairwise(iterable: Sequence[Any]) -> Iterator[tuple[Any, Any]]:
    """
    Iterate over a sequence pairwise.

    Args:
        iterable:
            The sequence.

    Yields:
        The pairs.
    """
    try:
        return itertools.pairwise(iterable)
    except AttributeError:
        a, b = itertools.tee(iterable)
        next(b, None)
        return zip(a, b)


def parse_logging_level(logging_level: int | str) -> int:
    err_msg = (
        "logging_level must be one of "
        "loggin.DEBUG, logging.INFO, logging.WARNING, loggin.WARN, logging.ERROR, "
        "logging.CRITICAL, logging.FATAL, "
        "'DEBUG', 'INFO', 'WARNING', 'WARN', 'ERROR', 'CRITICAL', 'FATAL' "
        "'debug', 'info', 'warning', 'warn', 'error', 'critical', 'fatal',"
        "10, 20, 30, 40, 50."
    )
    if isinstance(logging_level, int):
        assert logging_level in [
            logging.DEBUG,
            logging.INFO,
            logging.WARNING,
            logging.ERROR,
            logging.FATAL
        ], err_msg
        return logging_level
    if isinstance(logging_level, str):
        assert logging_level in [
            "DEBUG", "INFO", "WARNING", "WARN", "ERROR", "CRITICAL", "FATAL",
            "debug", "info", "warnings", "warn", "error", "critical", "fatal"
        ], err_msg
        return getattr(  # type: ignore[no-any-return]
            logging, logging_level.upper()
        )
    else:
        raise TypeError("logging_level must be an int or a str.")



from __future__ import annotations

import copy
import logging
from collections.abc import Sequence
from pathlib import Path
from typing import Any

import torch
from torch import nn
from tqdm import tqdm

from rebasin.permutation_coordinate_descent import PermutationCoordinateDescent
from rebasin.utils import parse_logging_level


class MergeMany:
    """
    Implements the MergeMany algorithm from the Git Re-basin paper.

    Args:
        models:
            The models to merge.

            **ATTENTION**: The models are modified in-place! Make sure to pass
            copies of the models if you want to keep the originals.

            *Type*: :code:`Sequence[nn.Module | str | Path]`

        working_model:
            The model to use as the working model.
            This is necessary, because the model-architecture must be known.
            This is not the case for models loaded from a file.
            **This will also be modified in-place.**

            *Type*: :code:`nn.Module`

        input_data:
            The input data to use for the models.
            Can be anything; will be passed to :class:`PermutationCoordinateDescent`'s
            :code:`input_data_b` argument. Assumed to be on the same device as
            :code:`working_model`.

            *Type*: :code:`Any`

        device:
            The device to use for the models.

    Properties:
        merged_model:
            The merged model.
            This is only available after calling :py:meth:`~MergeMany.run`.
            Before, it raises an :code:`AttributeError`.

            *Type*: :code:`nn.Module`

    """

    def __init__(
            self,
            models: Sequence[nn.Module | str | Path],
            working_model: nn.Module,
            input_data: Any,
            device: torch.device | str | None = None,
            logging_level: str | int = logging.ERROR,
    ) -> None:
        self.models = models
        self.working_model = working_model
        self.input_data = input_data
        self.device = device
        self.logging_level = parse_logging_level(logging_level)

        self._merged_model: nn.Module = nn.Sequential()
        self._merged_model_was_caclulated = False

    @property
    def merged_model(self) -> nn.Module:
        """
        The merged model.
        """
        if not self._merged_model_was_caclulated:
            raise AttributeError(
                "The merged model has not been calculated yet. "
                "Please call MergeMany.run() first."
            )
        return self._merged_model

    def run(
            self, max_iterations: int = 100, max_iterations_pcd: int = 100
    ) -> nn.Module:
        """
        Runs the MergeMany algorithm.

        Args:
            max_iterations:
                The maximum number of iterations to run the MergeMany algorithm for.

                *Type*: :code:`int`

                *Default*: :code:`100`

            max_iterations_pcd:
                The :code:`max_iterations` argument to pass to
                the :class:`PermutationCoordinateDescent`'s
                :py:meth:`~PermutationCoordinateDescent.rebasin`-method.

        Returns:
            The merged model.

        Note:
            If the model contains any :code:`BatchNorm`-layers, it is recommended
            to run :py:func:`~rebasin.utils.recalculate_batch_norms` on the final
            model to recalculate the running stats.
        """
        if self.logging_level <= logging.INFO:
            print("Running MergeMany algorithm...")

        loop = tqdm(range(max_iterations), disable=self.logging_level > logging.INFO)
        for i in loop:
            progress = self._run_iteration(max_iterations_pcd, loop)
            if not progress:
                if self.logging_level <= logging.INFO:
                    print(f"Converged. Stopping early after {i} steps.")
                break

        self._merged_model = self._get_mean_model()
        self._merged_model_was_caclulated = True
        return self.merged_model


    def _run_iteration(self, max_iterations: int, loop: tqdm[Any]) -> bool:
        progress = False
        for i in torch.randperm(len(self.models)):
            mean_model = self._get_mean_model(except_index=i)
            self._set_working_model_state_dict(index=i)

            if self.logging_level <= logging.INFO:
                loop.set_description(f"Running PCD for model {i}...")

            pcd = PermutationCoordinateDescent(
                model_a=mean_model,
                model_b=self.working_model,
                input_data_b=self.input_data,
                device_a=self.device,
                device_b=self.device,
            )
            pcd.rebasin(max_iterations=max_iterations)

            self._store_working_model(index=i)
            progress = progress or self._calculate_progress(pcd)

        return progress

    @torch.no_grad()
    def _get_mean_model(self, except_index: int = -1) -> nn.Module:
        mean_model = copy.deepcopy(self.working_model)
        first_index = 0 if except_index != 0 else 1

        for i, model in enumerate(self.models):
            if i == except_index:
                continue

            if i == first_index:
                mean_model.load_state_dict(
                    model.state_dict()
                    if isinstance(model, nn.Module)
                    else torch.load(model)
                )
                continue

            # It is fine to use self.working_model here, because it is
            #   set again to the correct values after this method is called.
            self.working_model.load_state_dict(
                model.state_dict()
                if isinstance(model, nn.Module)
                else torch.load(model)
            )

            # Merge all the parameters.
            for param, mean_param in zip(
                    self.working_model.parameters(), mean_model.parameters()
            ):
                mean_param.data.add_(param.data)

        # Divide by the number of models.
        for param in mean_model.parameters():
            param.data.div_(
                len(self.models) - 1
                if 0 <= except_index < len(self.models)
                else len(self.models)
            )

        return mean_model

    def _set_working_model_state_dict(self, index: int) -> None:
        load_model = self.models[index]
        if isinstance(load_model, (str, Path)):
            self.working_model.load_state_dict(torch.load(load_model))
        else:
            self.working_model.load_state_dict(copy.deepcopy(load_model.state_dict()))

    def _store_working_model(self, index: int) -> None:
        store_model = self.models[index]
        if isinstance(store_model, (str, Path)):
            torch.save(self.working_model.state_dict(), store_model)
        else:
            store_model.load_state_dict(copy.deepcopy(self.working_model.state_dict()))

    @staticmethod
    def _calculate_progress(pcd: PermutationCoordinateDescent) -> bool:
        # If PermutationCoordinateDescent did not find any permutation that is not
        #   the identity permutation, then we have made no progress.
        # Otherwise, we have.
        return not all(
            torch.allclose(perm.perm_indices, torch.arange(len(perm.perm_indices)))
            for perm, _ in pcd.pinit.model_graph.permutation_to_info
        )
